/*
 * DYNABUF.C - contains routines for dynamic buffers.
 *
 * $Header:   C:/FE/SHARED/VCS/DYNABUF.C_V   1.2   03 Sep 1989 23:10:00   rji  $
 *
 * $Log:   C:/FE/SHARED/VCS/DYNABUF.C_V  $
 * 
 *    Rev 1.2   03 Sep 1989 23:10:00   rji
 * Changed #include file specifiers from <...> to "..." for non-standard
 * headers.
 * 
 *    Rev 1.1   21 Aug 1989 18:47:52   rji
 * Changed dbappend() return type to size_t from int (since could have buffers
 * larger than 32K).  Also, used size_t casts on pointer difference expressions
 * for type compatibility -- WARNING: ptrdiff_t is int in MSC, so a positive
 * difference of > 32K will be treated as a negative.
 * 
 *    Rev 1.0   07 Dec 1988 23:09:40   rji
 * Initial revision.
 * 
*/

/*=========================================================================*/
/*=                                                                       =*/
/*=      I N C L U D E S,  D E F I N E S,  D E C L A R A T I O N S        =*/
/*=                                                                       =*/
/*=========================================================================*/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         DYNABUF.C Includes

  $PATHS$
       Functions\C
       Shared\Includes
       Shared\Utils
       Modules\$./FL$
   $3$

   See Also:
        $/SEE(DEBUG.H Macros)$
        $/SEE(COMMON.H Macros)$
        $/SEE(DALLOC.H Macros)$
        $/SEE(DYNABUF.H Macros)$

   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* INCLUDES                                                                */
/*-------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "debug.h"
#include "common.h"
#include "dalloc.h"

#include "dynabuf.h"
/*                                                                   $end$ */


/*=========================================================================*/
/*=                                                                       =*/
/*=                F U N C T I O N   D E F I N I T I O N S                =*/
/*=                                                                       =*/
/*=========================================================================*/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         dbcreate()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DBCREATE                                                                */
/*                                                                         */
/* Create a dynamic buffer, i.e., a block of storage that can              */
/* expand beyond its originally allocated size to accommodate data that    */
/* otherwise would not fit.                                                */
/*                                                                         */
/* RETURNS                                                                 */
/*    Pointer to newly created dynamic buffer.  If create fails, error     */
/*    message is written to stderr and program exits.                      */
/*-------------------------------------------------------------------------*/

DYNA_BUFFER dbcreate( max_len, len_inc )
   size_t   max_len;
   size_t   len_inc;
/*                                                                   $end$ */
{
   DYNA_BUFFER buf;

   if ( max_len == 0 )
   {
      fprintf( stderr, "create_dyna_buffer: must specify max len\n" );
      exit( 1 );
   }

   if ( len_inc == 0 )
      len_inc = max_len / 2;                    /* default size increment  */

   if ( len_inc == 0 )
   {
      fprintf( stderr, "create_dyna_buffer: must specify len inc\n" );
      exit( 1 );
   }

   if ( ((buf = d_calloc( sizeof(DYNA_BUFFER_STRUCT), sizeof(char) )) == NULL)
        || ((buf->beg = d_calloc( max_len, sizeof(char) )) == NULL) )
   {
      fprintf( stderr, "create_dyna_buffer: unable to allocate memory\n" );
      exit( 1 );
   }

   /* buf->len and buffer storage area initialized to zeros by d_calloc()  */

   buf->end       = buf->beg;
   buf->max_len   = max_len;
   buf->len_inc   = len_inc;

   return ( buf );

} /* dbcreate() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         dbappend()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DBAPPEND                                                                */
/*                                                                         */
/* Add data to a dynamic buffer starting at the next available byte of the */
/* buffer.  If the data is too long to fit in the space left, the buffer   */
/* is expanded so that the data can be accommodated.                       */
/*                                                                         */
/* RETURNS                                                                 */
/*    offset to data stored.                                               */
/*-------------------------------------------------------------------------*/

size_t   dbappend( buf, data, data_len )
   REG1 DYNA_BUFFER     buf;
   void                *data;
   size_t               data_len;
/*                                                                   $end$ */
{
   char              *p;
   size_t            new_size;

   if ( data_len == 0 )
      return ( 0 );

   /* Check for valid buffer */

   if ( buf == NULL )
   {
      fprintf( stderr, "append_dyna_buffer: error - NULL buffer pointer\n" );
      exit( 1 ) ;
   }

   if ( (size_t) (buf->end - buf->beg) != buf->len )
   {
      fprintf( stderr, "append_dyna_buffer: error - buffer messed up\n" );
      exit( 1 ) ;
   }

   /* Re-size buffer, if necessary */

   if ( buf->len + data_len >= buf->max_len )
   {
      /* expand dynamic buffer by the minimal # of length increments to
      ** accommodate data while still leaving extra space for more data
      */
      new_size = (((buf->len + data_len - buf->max_len) / buf->len_inc) + 1)
                 * buf->len_inc + buf->max_len;
      if ( (p = (char *) d_realloc( buf->beg, new_size )) == NULL )
      {
         fprintf( stderr, "append_dyna_buffer: Could not reallocate buf,"
                          " size requested: %d\n", new_size );
         exit( 1 );
      }
      buf->max_len = new_size;
      buf->beg = p;
      buf->end = buf->beg + buf->len;
   }

   /* Append data */

   memcpy( buf->end, data, data_len );          /* Append data */
   buf->end    += data_len;
   buf->len    += data_len;
   *buf->end   =  '\0';                      /* Make sure null is at end   */

   /* Return offset to data stored */

   return ( buf->len - data_len );

} /* dbappend() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         dbclear()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DBCLEAR                                                                 */
/*                                                                         */
/* Resets dynamic buffer control variables, so buffer appears as though    */
/* freshly created with it's current max len and len increment.  No memory */
/* allocation or deallocation is involved.                                 */
/*-------------------------------------------------------------------------*/

int   dbclear( buf )
   REG1 DYNA_BUFFER  buf;
/*                                                                   $end$ */
{
   /* Check for valid buffer */

   if ( buf == NULL )
   {
      fprintf( stderr, "clear_dyna_buffer: NULL buffer pointer\n" );
      exit( 1 );
   }

   if ( (size_t) (buf->end - buf->beg) != buf->len )
   {
      fprintf( stderr, "clear_dyna_buffer: error - buffer messed up\n" );
      exit( 1 ) ;
   }

   /* Clear buffer */

   *buf->beg      = '\0';
   buf->end       = buf->beg;
   buf->len       = 0;

   return ( 1 );

} /* dbclear() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         dbdestroy()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DBDESTROY                                                               */
/*                                                                         */
/* Frees all memory allocated to dyna buffer                               */
/*-------------------------------------------------------------------------*/

void  dbdestroy( bp )
   DYNA_BUFFER *bp;
/*                                                                   $end$ */
{

   /* Check for valid buffer */

   if ( *bp == NULL )
   {
      fprintf( stderr, "destroy_dyna_buffer: NULL buffer pointer\n" );
      exit( 1 );
   }

   if ( (size_t) ((*bp)->end - (*bp)->beg) != (*bp)->len )
   {
      fprintf( stderr, "destroy_dyna_buffer: error - buffer messed up\n" );
      exit( 1 ) ;
   }

   /* Free data storage area */

   if ( (*bp)->beg != NULL )
      d_free( (*bp)->beg );

   /* Free header structure & null out pointer to prevent further access */

   d_free( *bp );
   *bp = NULL;

} /* dbdestroy() */
